#!/usr/bin/perl

# used https://www.codeconvert.ai/php-to-perl-converter to convert frame-buffer-lister.php

use strict;
use warnings;
use File::Glob ':glob';
use POSIX qw(strftime);

sub padHex {
    my ($d, $l) = @_;
    return sprintf("%0${l}x", $d);
}

sub toHex {
    my ($s) = @_;
    my $t = '';
    foreach my $char (split //, $s) {
        $t .= padHex(ord($char), 2);
    }
    return $t;
}

sub toStr {
    my ($h) = @_;
    my $s = '';
    for (my $i = 4; $i-- > 0;) {
        $s .= chr(($h >> (8 * $i)) & 0xFF);
    }
    return $s;
}

sub myread {
    my ($fh, $len) = @_;
    my $buffer;
    read($fh, $buffer, $len);
    return $buffer;
}

my %ctype = (
    '02000000' => 'LVDS',
    '04000000' => 'DVI-D',
    '08000000' => 'SVIDEO',
    '10000000' => 'VGA',
    '00020000' => 'DVI-SL',
    '00040000' => 'DP',
    '00080000' => 'HDMI',
    '00100000' => 'UNKNOWN',
    '80000000' => 'S-V',
);

#printf("hello\n");
#foreach my $file (glob("$dir*Controller.kext")) 
my $file = $ARGV[0];
{
    my $binary_file = $file;
    #printf("thefile:%s]\n", $binary_file);
    #open my $a, '-|', "otool -aSfhlLorcITRMHGCPvtdjV $binary_file" or die $!;
    open my $a, '-|', "otool -ljVt $binary_file" or die $!;
    open my $b, '<', $binary_file or die $!;

    my $line=0;
	my $sectname = "";
	my $segname = "";
	my $dataaddr = 0;
	my $datasize = 0;
	my $dataoffset = 0;
	my $is32 = 0;
    while (my $l = <$a>) {
    	$line += 1;
    	if ($l =~ /\bcmd LC_SEGMENT\b/) {
    		$is32 = 1;
    	}
    	if ($l =~ /\bsectname (\w+)\b/) {
    		$sectname = $1;
    		#printf("sectname «%s» isdata:%d\n", $sectname, ($sectname eq "__data"));
    	}
    	if ($l =~ /\bsegname (\w+)\b/) {
    		$segname = $1;
    		#printf("segname «%s» isdata:%d\n", $segname, ($segname eq "__DATA"));
    	}
    	if (($sectname eq "__data") && ($segname eq "__DATA") && ($l =~ /\baddr (0x\w+)\b/)) {
    		$dataaddr = hex($1);
    		#printf("dataaddr «0x%x»\n", $dataaddr);
    	}
    	if (($sectname eq "__data") && ($segname eq "__DATA") && ($l =~ /\boffset (\d+)\b/)) {
    		$dataoffset = $1;
    		#printf("dataoffset «0x%x»\n", $dataoffset);
    	}
    	
        if ($l =~ /^__ZN\d+(\w+)Info10createInfo.*/) {
			#printf("• %d %s", $line, $l);
			my $f = {};
			$f->{name} = $1;
			$f->{size} = 16;
	
			while ($l = <$a>) {
				$line += 1;
				if ($l =~ /\bret[lq]\b/) {
					last;
				}
				#printf("• %d %s", $line, $l);
				if ($l =~ /(\w+)\t(\w{6}(\w\w)(\w\w)(\w\w)(\w\w))\s+leaq\t__ZL10CONNECTORS\(%rip\)/) {
					$f->{addr} = hex($1) + hex($6 . $5 . $4 . $3) + (length($2)/2);
					#printf("° %d =0x%x: %s", $line, $f->{addr}, $l);
				}
				if ($l =~ /(\w+)\t(\w{6}(\w\w)(\w\w)(\w\w)(\w\w))\s+leaq\t__ZL18CONNECTORS_CONFIGS\(%rip\)/) {
					$f->{portsaddr} = hex($1) + hex($6 . $5 . $4 . $3) + (length($2)/2);
					#printf("° %d =0x%x: %s", $line, $f->{addr}, $l);
				}
				if ($is32 && $l =~ /\w+\t\w{4}(\w\w)(\w\w)(\w\w)(\w\w)\s+leal\t__ZL10CONNECTORS\(%e\w+\)/) {
					$f->{addr} = hex($4 . $3 . $2 . $1);
					#printf("° %d =0x%x: %s", $line, $f->{addr}, $l);
				}
				if ($is32 && $l =~ /\w+\t\w{2,6}(\w\w)(\w\w)(\w\w)(\w\w)\s+movl\t\$__ZL10CONNECTORS,/) {
					$f->{addr} = hex($4 . $3 . $2 . $1);
					#printf("° %d =0x%x: %s", $line, $f->{addr}, $l);
				}
				if ($is32 && $l =~ /\w+\t\w\w(\w\w)(\w\w)(\w\w)(\w\w)\s+addl\t\$__ZL10CONNECTORS,/) {
					$f->{addr} = hex($4 . $3 . $2 . $1);
					#printf("° %d =0x%x: %s", $line, $f->{addr}, $l);
				}
				if ($l =~ /movb\t\$(0x\w+), 0x[13]\b/) {
					$f->{ports} = hex($1) & 0xFF;
					#printf("° %d =0x%x: %s", $line, $f->{ports}, $l);
				}
				if ($is32 && $l =~ /\w+\t\w{6}(\w\w)(\w\w)(\w\w)(\w\w)\s+movzbl\t\__ZL18CONNECTORS_CONFIGS\(/) {
					$f->{portsaddr} = hex($4 . $3 . $2 . $1);
					#printf("° %d =0x%x: %s", $line, $f->{ports}, $l);
				}
				if ($l =~ /shlq\t\$(0x\w+),/) {
					$f->{size} = 1 << hex($1);
					#printf("° %d =0x%x: %s", $line, $f->{ports}, $l);
				}
				if ($is32 && $l =~ /shll\t\$(0x\w+),/) {
					$f->{size} = 1 << hex($1);
					#printf("° %d =0x%x: %s", $line, $f->{ports}, $l);
				}
				if ($l =~ /imulq\t\$(0x\w{2,3}),/) {
					$f->{size} = hex($1);
					#printf("° %d =0x%x: %s", $line, $f->{ports}, $l);
				}
				if ($l =~ /addq\t\$(0x\w{2,3}), %[re][acd]x/) {
					$f->{size} = hex($1);
					#printf("° %d =0x%x: %s", $line, $f->{ports}, $l);
				}
				if ($is32 && $l =~ /addl\t\$(0x\w{2,3}), %[re][acd]x/) {
					$f->{size} = hex($1);
					#printf("° %d =0x%x: %s", $line, $f->{ports}, $l);
				}
			}
			#printf("%d", $line);
			printf("%s", $f->{name});
			if (exists $f->{portsaddr}) {
				seek($b, $f->{portsaddr} - $dataaddr + $dataoffset, 0);
				$f->{ports} = hex(toHex(myread($b, 1)));
			}
			if (exists $f->{ports}) {
				printf(" (%d)", $f->{ports});
			}
			else {
				printf(" (no ports)");
			}
			printf(" @ 0x%x (0x%x)\n", $f->{addr}, $f->{addr} - $dataaddr + $dataoffset);
	
			my @t;
			my @p;
			if (exists $f->{ports}) {
				seek($b, $f->{addr} - $dataaddr + $dataoffset, 0);
				while ($f->{ports}-- > 0) {
					my $thebytes = myread($b, $f->{size});
					my $thehex = toHex($thebytes);
					my $thetype = substr($thehex, 0, 8);
					my $porttype = $ctype{$thetype};
					push @t, $thehex;
					push @p, $porttype;
				}
				print join(', ', @p), "\n";
				print join("\n", @t), "\n\n";
			}
		}
    }
    close $a;
    close $b;
}
